@@ -2,10 +2,14 @@
 
 from opentps.core._event import Event
 from opentps.core._api import APIInterpreter
+import opentps.core._loggingConfig as loggingConfig
 
 import opentps.core.data as data
 import opentps.core.io as io
 import opentps.core.processing as processing
 import opentps.core.utils as utils
+import opentps.core.examples as examples
+
+loggingConfig.configure()
 
 __all__ = [s for s in dir() if not s.startswith('_')]
